#################### Imports ####################
# Standard libraries
import asyncio
import threading
import seaborn as sns
import sys

# IDUN sdk
from idun_guardian_sdk import FileTypes, GuardianClient

# Local libraries
from front_end.EEGVisualizer import *


#################### Configuration ####################
####################   TO MODIFY   ####################
# Recording settings
RECORDING_TIMER = int(60 * 30) # 30 minutes
LED_SLEEP = False

# Device ID
my_api_token = "INSERT_HERE" 
device_address = "INSERT_HERE" 
#############   END OF MODIFICATIONS   ################



################## Initialize variables and visualizer ###########
# Classifiers 
fft_keys = ['Delta', 'Theta', 'Alpha', 'Sigma', 'Beta', 'Gamma']

# Visualizer
app = QtWidgets.QApplication(sys.argv)
visualizer = EEGVisualizer(plot_FFT=True)
sns.set(style="whitegrid")


################### Handler functions ####################
def signal_handler(event):
    """
    Handler for the live signal data (EEG and IMU)
    """
    global visualizer

    # EEG Signal
    data = event.message
    eeg_samples = [sample['ch1'] for sample in data['filtered_eeg']]

    # IMU data
    imu_data = data['imu']

    # Update the visualizer
    visualizer.update_signals_visuals(eeg_samples, imu_data)

                                
def pred_handler(event):
    """
    Handler for the live prediction data (Jaw clench, HEOG, FFT)
    """
    global visualizer

    # Initialize the classifiers values
    jaw_clench = False
    heog_left = False
    heog_right = False
    fft_dict = {}

    # Jaw clench 
    if any('predictionResponse' in key for key in event.message):
        jaw_clenches = event.message["predictionResponse"]
        binary_jaw_clenches = [0 if jaw_clench == 'Nothing' else 1 for jaw_clench in jaw_clenches]

        if any([pred > 0 for pred in binary_jaw_clenches]):
            jaw_clench = True

    # HEOG 
    if any("heogClassificationsResponse" in key for key in event.message):
        heog = event.message["heogClassificationsResponse"]

        if any([pred == 1 for pred in heog]):
            heog_left = True

        if any([pred == 2 for pred in heog]):
            heog_right = True

    # FFT
    if any("stateless_z_scores" in key for key in event.message):
        if len(event.message["stateless_z_scores"]) == 0:
            fft_dict = {key: None for key in fft_keys}
        else:
            fft_dict = event.message["stateless_z_scores"][0]

    # Update the visualizer
    visualizer.update_predictions_visuals(jaw_clench, heog_left, heog_right, fft_dict)
        

#################### Recording ####################
recording_done = False

async def do_recording(client):
    """
    Asynchronous function to handle the recording process.
    """
    global recording_done

    # Start the recording
    await client.start_recording(recording_timer=RECORDING_TIMER, led_sleep=LED_SLEEP, calc_latency=False)

    # Other actions after recording ends
    rec_id = client.get_recording_id()
    print("RecordingId", rec_id)
    client.update_recording_tags(recording_id=rec_id, tags=["tag1", "tag2"])
    client.update_recording_display_name(recording_id=rec_id, display_name="todays_recordings")
    client.download_file(recording_id=rec_id, file_type=FileTypes.EEG)

    # Close the app
    recording_done = True

def check_for_completion():
    """ 
    Periodically check if the recording is done and close the app if so. 
    """
    global app, visualizer
    if recording_done:
        visualizer.close()  
        app.quit()  
    else:
        QtCore.QTimer.singleShot(100, check_for_completion)


#################### MAIN ##########################
if __name__ == "__main__":
    # Start the real-time visualization
    visualizer.show()

    # Create the client and subscribe to live insights
    client = GuardianClient(api_token=my_api_token, address=device_address)
    client.subscribe_live_insights(raw_eeg=False, filtered_eeg=True, imu=True, handler=signal_handler)
    client.subscribe_realtime_predictions(jaw_clench=True, fft=True, bin_heog=True, handler=pred_handler)

    # Start a separate thread for the async recording
    recording_thread = threading.Thread(target=lambda: asyncio.run(do_recording(client)), daemon=True)
    recording_thread.start()

    # Periodically check if the recording is done and close the app if so
    QtCore.QTimer.singleShot(100, check_for_completion)

    # Exit the app
    sys.exit(app.exec())
