import numpy as np
from PyQt6 import QtWidgets, QtCore
from PyQt6.QtGui import QPixmap
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from collections import deque

window_size = 30
window_size_fft = 15
window_size_signal = window_size * 20 * 7  
window_size_imu = window_size_signal // 10
image_size = 100
hold_duration = 500

brainwave_bands = {
    'Delta': (0.5, 4),
    'Theta': (4, 8),
    'Alpha': (8, 12),
    'Sigma': (12, 15),
    'Beta': (15, 30),
    'Gamma': (30, 35)
}

band_colors = {
    'Delta': '#A4C2F4',    # Pastel blue
    'Theta': '#76D7C4',    # Pastel green
    'Alpha': '#B6D7A8',    # Pastel turquoise
    'Sigma': '#F9CB9C',    # Pastel peach
    'Beta': '#F6A5A5',     # Pastel red
    'Gamma': '#DDA0DD'     # Pastel purple
}

imu_colors = {
    'x': 'r',   # Red
    'y': 'g',   # Green
    'z': 'b'    # Blue
}

class EEGVisualizer(QtWidgets.QWidget):
    """
    Class for real-time visualization of EEG data.

    Attributes:
        plot_FFT (bool): Whether to plot the FFT data.
        signal_data (deque): Buffer for EEG signal data.
        fft_keys (list): Keys for the FFT data.
        fft_dict (dict): Buffer for FFT data.
        imu_keys (list): Keys for the IMU data.
        imu_dict (dict): Buffer for IMU data.
        timer (QTimer): Timer for updating the visualization.
        jaw_clench_detected (bool): Whether a jaw clench was detected.
        heog_left_detected (bool): Whether a left HEOG was detected.
        heog_right_detected (bool): Whether a right HEOG was detected.
        jaw_clench_hold_timer (int): Timer for holding the jaw clench detection.
        heog_left_hold_timer (int): Timer for holding the left HEOG detection.
        heog_right_hold_timer (int): Timer for holding the right HEOG detection.

    """
    def __init__(self, plot_FFT=False):
        """
        Initialize the EEGVisualizer object.

        Args:
            plot_FFT (bool): Whether to plot the FFT data.
        """
        super().__init__()
        
        # Parameters
        self.plot_FFT = plot_FFT
        
        # Initialize variables for real-time data
        self.signal_data = deque(maxlen=window_size_signal)  # buffer for EEG signal data
        self.signal_data.extend([0 for _ in range(window_size_signal)])
        self.jaw_clench_detected = False
        self.heog_left_detected = False
        self.heog_right_detected = False

        # FFT data
        if plot_FFT:
            self.fft_keys = ['Delta', 'Theta', 'Alpha', 'Beta', 'Gamma']
            self.fft_dict = {key: deque(maxlen=window_size_fft) for key in self.fft_keys}
            for key in self.fft_keys:
                self.fft_dict[key].extend([0 for _ in range(window_size_fft)])

        # IMU data
        self.acc_keys = ['acc_x', 'acc_y', 'acc_z']
        self.magn_keys = ['magn_x', 'magn_y', 'magn_z']
        self.gyro_keys = ['gyro_x', 'gyro_y', 'gyro_z']
        self.imu_keys = self.acc_keys + self.magn_keys + self.gyro_keys
        self.imu_dict = {key: deque(maxlen=window_size_imu) for key in self.imu_keys}
        for key in self.imu_keys:
            self.imu_dict[key].extend([0 for _ in range(window_size_imu)])

        # Set up timer for updating the visualization
        self.timer = QtCore.QTimer()
        self.timer.setInterval(50)  # Update interval in milliseconds
        self.timer.timeout.connect(self.update_plot)
        self.timer.start()

        # Timers to keep track of when to reset each detection
        self.jaw_clench_hold_timer = 0
        self.heog_left_hold_timer = 0
        self.heog_right_hold_timer = 0

        self.initUI()

    def initUI(self):
        """
        Set up the user interface.
        """
        self.setStyleSheet("background-color: white; color: black;")

        main_layout = QtWidgets.QHBoxLayout()  # Main horizontal layout for the entire window
        
        # Set a title for left and right sections
        title_style_sections = """
            QLabel {
                font-size: 20px;    /* Change font size */
                font-weight: bold;  /* Make the text bold */
                color: #404040;     /* Change text color if needed */
            }
        """

        #---------- LEFT: EEG ----------------------------------------------------------------
        left_layout = QtWidgets.QVBoxLayout()  # Vertical layout for the left side of the window

        title_signal = QtWidgets.QLabel("EEG Data")
        title_signal.setStyleSheet(title_style_sections)
        title_signal.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        # Add title label to layout
        left_layout.addWidget(title_signal)
        
        #---------- Horizontal layout to hold the time-series plots 
        top_layout = QtWidgets.QHBoxLayout()
        plot_layout = QtWidgets.QVBoxLayout()  

        if self.plot_FFT:
            self.figure, (self.ax_signal, self.ax_fft) = plt.subplots(2, 1, figsize=(12, 12))
        else:
            self.figure, self.ax_signal = plt.subplots(1, 1, figsize=(12, 8))

        self.canvas = FigureCanvas(self.figure)
        self.figure.tight_layout(pad=5.0)

        # Filtered signal 
        self.ax_signal.set_title("Filtered EEG Signal (µV)")
        self.line_signal, = self.ax_signal.plot(self.signal_data, color='#14786e', linewidth=1)
        self.ax_signal.set_ylim(-100, 100)
        self.ax_signal.set_xticks([])

        # FFT plot
        if self.plot_FFT:
            self.ax_fft.set_title("FFT (z-scores)")
            self.fft_lines = {}
            for key in self.fft_keys:
                self.fft_lines[key], = self.ax_fft.plot(self.fft_dict[key], label=f"{key} {brainwave_bands[key]}", color=band_colors[key],linewidth=1.5)
            self.ax_fft.set_ylim(-10, 10)
            self.ax_fft.set_xticks([])
            self.ax_fft.legend(loc='upper left')

        # Add canvas with the plots to the plot layout
        plot_layout.addWidget(self.canvas)
        top_layout.addLayout(plot_layout)
        left_layout.addLayout(top_layout)

        #------------ Classifiers layout 
        # Create a grid layout for the images (jaw clench, left and right HEOG images)
        image_layout = QtWidgets.QGridLayout()

        # Titles
        title_style = """
            QLabel {
                font-size: 20px;    /* Change font size */
                font-weight: bold;  /* Make the text bold */
                color: #404040;     /* Change text color if needed */
            }
        """
        title_jaw_clench = QtWidgets.QLabel("Jaw Clench")
        title_eye_movements = QtWidgets.QLabel("Left HEOG")
        title_eye_movements_right = QtWidgets.QLabel("Right HEOG")
        title_jaw_clench.setStyleSheet(title_style)
        title_eye_movements.setStyleSheet(title_style)
        title_eye_movements_right.setStyleSheet(title_style)

        # Add title labels to layout
        image_layout.addWidget(title_eye_movements, 0, 0, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)
        image_layout.addWidget(title_jaw_clench, 0, 1, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)
        image_layout.addWidget(title_eye_movements_right, 0, 2, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)

        # Initialize image labels for jaw clench and HEOG states
        self.jaw_clench_image = QtWidgets.QLabel()
        self.heog_left_image = QtWidgets.QLabel()
        self.heog_right_image = QtWidgets.QLabel()

        # Set default images (resting state)
        resting_jaw_clench_image = QPixmap("front_end/images/circle_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)
        resting_heog_left_image = QPixmap("front_end/images/left_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)
        resting_heog_right_image = QPixmap("front_end/images/right_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio, QtCore.Qt.TransformationMode.SmoothTransformation)

        self.jaw_clench_image.setPixmap(resting_jaw_clench_image)
        self.heog_left_image.setPixmap(resting_heog_left_image)
        self.heog_right_image.setPixmap(resting_heog_right_image)

        # Add images to layout
        image_layout.addWidget(self.heog_left_image, 1, 0, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)
        image_layout.addWidget(self.jaw_clench_image, 1, 1, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)
        image_layout.addWidget(self.heog_right_image, 1, 2, alignment=QtCore.Qt.AlignmentFlag.AlignCenter)

        # Add image layout to the main layout (below the plots)
        left_layout.addLayout(image_layout)

        #--------- Add left layout to the main layout
        left_layout.setContentsMargins(20, 20, 20, 20)
        main_layout.addLayout(left_layout)

        #---------- RIGHT: IMU ----------------------------------------------------------------
        right_layout = QtWidgets.QVBoxLayout()  # Vertical layout for the right side of the window

        # Title for the IMU section
        title_imu = QtWidgets.QLabel("IMU Data")
        title_imu.setStyleSheet(title_style_sections)
        title_imu.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        # Add title label to layout
        right_layout.addWidget(title_imu)

        # Create a figure and axis for the IMU data
        self.figure_imu, (self.ax_acc, self.ax_magn, self.ax_gyro) = plt.subplots(3, 1, figsize=(12, 12))
        self.canvas_imu = FigureCanvas(self.figure_imu)
        self.figure_imu.tight_layout(pad=3.0)

        # Accelerometer plot
        self.ax_acc.set_title("Accelerometer Data (m/s²)")
        self.acc_lines = {}
        for key in self.acc_keys:
            axis = key.split('_')[1]
            self.acc_lines[key], = self.ax_acc.plot(self.imu_dict[key], label=axis, linewidth=1, color=imu_colors[axis])
        self.ax_acc.set_ylim(-20, 15)
        self.ax_acc.set_xticks([])
        self.ax_acc.legend(loc='upper left')

        # Magnetometer plot
        self.ax_magn.set_title("Magnetometer Data (µT)")
        self.magn_lines = {}
        for key in self.magn_keys:
            axis = key.split('_')[1]
            self.magn_lines[key], = self.ax_magn.plot(self.imu_dict[key], label=axis, linewidth=1, color=imu_colors[axis])
        self.ax_magn.set_ylim(-2, 2)
        self.ax_magn.set_xticks([])
        self.ax_magn.legend(loc='upper left')

        # Gyroscope plot
        self.ax_gyro.set_title("Gyroscope Data (°/s)")
        self.gyro_lines = {}
        for key in self.gyro_keys:
            axis = key.split('_')[1]
            self.gyro_lines[key], = self.ax_gyro.plot(self.imu_dict[key], label=axis, linewidth=1, color=imu_colors[axis])
        self.ax_gyro.set_ylim(-7, 7)
        self.ax_gyro.set_xticks([])
        self.ax_gyro.legend(loc='upper left')

        # Add canvas with the plots to the right layout
        right_layout.addWidget(self.canvas_imu)
        right_layout.setContentsMargins(20, 20, 20, 20)
        main_layout.addLayout(right_layout)

        #-------------------------
        # Set the main layout for the window
        self.setLayout(main_layout)

    
    def update_signals_visuals(self, signal, imu_data):
        """
        Update the EEG signal and IMU data buffers using live insights.

        Args:
            signal (list): New EEG signal data.
            imu_data (dict): New IMU data.
        """
        # Update EEG data
        self.signal_data.extend(signal)

        # Update IMU data
        for sample in imu_data:
            for key in self.imu_keys:
                self.imu_dict[key].append(sample[key])
    
    def update_predictions_visuals(self, new_jaw_clench, new_heog_left, new_heog_right, fft_dict):
        """
        Update the EEG prediction data using real-time predictions.

        Args:
            new_jaw_clench (bool): New jaw clench detection.
            new_heog_left (bool): New left HEOG detection.
            new_heog_right (bool): New right HEOG detection.
            fft_dict (dict): New FFT z-scores data.
        """
        # Get current time for classifier hold timers
        current_time = QtCore.QTime.currentTime().msecsSinceStartOfDay()

        # Update detection hold timers
        if new_jaw_clench:
            self.jaw_clench_hold_timer = current_time + hold_duration
        elif current_time < self.jaw_clench_hold_timer:
            new_jaw_clench = True

        if new_heog_left:
            self.heog_left_hold_timer = current_time + hold_duration
        elif current_time < self.heog_left_hold_timer:
            new_heog_left = True

        if new_heog_right:
            self.heog_right_hold_timer = current_time + hold_duration
        elif current_time < self.heog_right_hold_timer:
            new_heog_right = True

        # Update detection flags
        self.jaw_clench_detected = new_jaw_clench
        self.heog_left_detected = new_heog_left
        self.heog_right_detected = new_heog_right

        # Update FFT data
        if fft_dict!={}:
            for key in self.fft_keys:
                if fft_dict[key] is not None:
                    self.fft_dict[key].append(fft_dict[key])
        

    def update_plot(self):
        """
        Update the real-time EEG plots
        Automatically called by the timer.
        """
        # Update EEG data
        self.line_signal.set_ydata(self.signal_data)
        if self.plot_FFT:
            for key in self.fft_keys:
                self.fft_lines[key].set_ydata(self.fft_dict[key])
        self.canvas.draw()

        # Update IMU data
        for key in self.acc_keys:
            self.acc_lines[key].set_ydata(self.imu_dict[key])
        for key in self.magn_keys:
            self.magn_lines[key].set_ydata(self.imu_dict[key])
        for key in self.gyro_keys:
            self.gyro_lines[key].set_ydata(self.imu_dict[key])
        self.canvas_imu.draw()

        # jaw clench
        if self.jaw_clench_detected:
            active_jaw_clench_image = QPixmap("front_end/images/jaw_clench.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.jaw_clench_image.setPixmap(active_jaw_clench_image)
        else:
            resting_jaw_clench_image = QPixmap("front_end/images/circle_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.jaw_clench_image.setPixmap(resting_jaw_clench_image)

        # HEOG 
        if self.heog_left_detected:
            active_heog_left_image = QPixmap("front_end/images/HEOG_left.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.heog_left_image.setPixmap(active_heog_left_image)
        else:
            resting_heog_left_image = QPixmap("front_end/images/left_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.heog_left_image.setPixmap(resting_heog_left_image)

        if self.heog_right_detected:
            active_heog_right_image = QPixmap("front_end/images/HEOG_right.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.heog_right_image.setPixmap(active_heog_right_image)
        else:
            resting_heog_right_image = QPixmap("front_end/images/right_grey.png").scaled(image_size, image_size, QtCore.Qt.AspectRatioMode.KeepAspectRatio,QtCore.Qt.TransformationMode.SmoothTransformation)
            self.heog_right_image.setPixmap(resting_heog_right_image)


